# Generated by Django 3.1.14 on 2022-11-16 14:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('whistle', '0005_events_settings'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='event',
            field=models.CharField(db_index=True, max_length=50),
        ),
    ]
